"""
Collection of functions for visualizing EIS data and results.

.. currentmodule:: autoeis.visualization

.. autosummary::
   :toctree: generated/

    draw_circuit
    plot_bode
    plot_nyquist
    plot_impedance_combo
    plot_linKK_residuals
    plot_nyquist
    set_plot_style
    show_nticks

"""

import logging
import os
import re
from collections.abc import Iterable

import arviz
import matplotlib as mpl
import matplotlib.colors as mcolors
import matplotlib.pyplot as plt
import numpy as np
import numpyro
import pandas as pd
import rich
import seaborn as sns
from matplotlib.axes import Axes
from pandas.io.formats.style import Styler
from rich.console import Console
from rich.table import Table

log = logging.getLogger(__name__)


__all__ = [
    "draw_circuit",
    "plot_bode",
    "plot_nyquist",
    "plot_impedance_combo",
    "plot_linKK_residuals",
    "plot_nyquist",
    "print_summary_statistics",
    "print_inference_results",
    "rich_print",
    "set_plot_style",
    "show_nticks",
]


def is_interactive_session() -> bool:  # pragma: no cover
    """Returns True if the code is running in an interactive session, False otherwise."""
    return hasattr(__builtins__, "__IPYTHON__")


def is_ipython_notebook() -> bool:  # pragma: no cover
    """Returns True if the code is running in a Jupyter notebook, False otherwise."""
    try:
        shell = get_ipython().__class__.__name__
        if shell == "ZMQInteractiveShell":
            return True  # Jupyter notebook or qtconsole
        if shell == "TerminalInteractiveShell":
            return False  # Terminal running IPython
        return False  # Other type (?)
    except NameError:
        return False  # Probably standard Python interpreter


def rich_print(*args, **kwargs):
    """Overrides the built-in print() function with rich's print() function."""
    if is_ipython_notebook():
        # HACK: prevent rich from creating a new <div> for every print()
        # NOTE: works for notebooks, but breaks for interactive sessions
        console = Console(force_jupyter=False)
        console.print(*args, **kwargs)
    else:
        rich.print(*args, **kwargs)


def draw_circuit(circuit: str) -> mpl.figure.Figure:
    """Draws the circuit model using lcapy.

    Parameters
    ----------
    circuit : str
        CDC string representation of the input circuit. See
        `here <https://autodial.github.io/AutoEIS/circuit.html>`_ for details.

    Returns
    -------
    fig : lcapy.figure
        Handle of the circuit figure
    """
    try:
        from lcapy import CPE as P
        from lcapy import C, L, R
    except ImportError:
        msg = "lcapy is not installed. Please install it using `pip install lcapy`."
        log.error(msg)
        return

    # Replace square brackets with parentheses
    circuit = circuit.replace("[", "(").replace("]", ")")
    # Replace commas with vertical bars
    circuit = circuit.replace(",", "|")
    # Replace dashes with plus signs
    circuit = circuit.replace("-", "+")
    # Surround all numbers with parentheses
    circuit = re.sub(r"([A-Z])(\d+)", r'\1("\1\2")', circuit)

    fig = eval(circuit)
    fig.draw(style="american")

    return fig


def plot_nyquist(
    Z: np.ndarray[complex],
    *,
    fmt: str = ".-",
    markersize: int = 6,
    color: str = None,
    alpha: float = 1,
    label: str = None,
    ax: plt.Axes = None,
) -> plt.Axes:
    """Plots EIS data in Nyquist plot.

    Parameters
    ----------
    Z: np.ndarray[complex]
        Impedance data.
    fmt: str, optional
        Format of the markers in the plot. Default is "o-".
    markersize: int, optional
        Size of the markers in the plot. Default is 6.
    color: str, optional
        Color of the markers in the plot. Default is None.
    alpha: float, optional
        Transparency of the markers in the plot. Default is 1.
    label: str, optional
        Label for the plot. Default is None.
    ax: plt.Axes, optional
        Axes to plot on. Default is None.

    Returns
    -------
    plt.Axes
        Axes object of the Nyquist plot.
    """
    if ax is None:
        fig, ax = plt.subplots(figsize=(4, 3), layout="constrained")

    # Remove color from fmt if present
    if fmt[0] in ["b", "g", "r", "c", "m", "y", "k", "w"]:
        color = fmt[0]
        fmt = fmt[1:]

    ax.plot(Z.real, -Z.imag, fmt, c=color, markersize=markersize, label=label, alpha=alpha)
    ax.set_xlabel("Re(Z)")
    ax.set_ylabel("-Im(Z)")
    ax.axis("equal")

    if label is not None:
        ax.legend()

    return ax


def plot_bode(
    freq: np.ndarray,
    Z: np.ndarray,
    *,
    fmt=".-",
    markersize=6,
    deg: bool = True,
    ax: np.ndarray[plt.Axes] = None,
    label: str = None,
    log: bool = False,
    alpha: float = 1,
) -> np.ndarray:
    """Plots the Bode plot for the impedance data.

    Parameters
    ----------
    freq: np.ndarray
        Frequencies corresponding to the impedance data.
    Z: np.ndarray
        Impedance data.
    fmt: str, optional
        Format of the markers in the plot. Default is ".-".
    markersize: int, optional
        Size of the markers in the plot. Default is 6.
    deg: bool, optional
        If True, plots the Bode plot in degrees. Default is True.
    ax: np.ndarray, optional
        Array of Axes to plot on. Default is None.
    label: str, optional
        Label for the data series. Default is None.
    log: bool, optional
        If True, plots the y-axis of |Z| vs. freq in log scale. Default is False.
    alpha: float, optional
        Transparency of the markers in the plot. Default is 1.

    Returns
    -------
    np.ndarray
        Array of Axes objects of the Bode plot.
    """
    if ax is None:
        fig, ax = plt.subplots(ncols=2, figsize=(8, 3.25), layout="constrained")

    if not isinstance(ax, np.ndarray) or not isinstance(ax[0], plt.Axes):
        raise AssertionError("Incompatible 'ax': Must be two-long subplot axes. "
                             "Use `fig, ax = plt.subplots(ncols=2)` and pass 'ax'.")  # fmt: off

    # Plot magnitude |Z|
    ax[0].plot(freq, np.abs(Z), fmt, label=label, markersize=markersize, alpha=alpha)
    ax[0].set_xscale("log")
    ax[0].set_xlabel("frequency (Hz)")
    ax[0].set_ylabel(r"$|Z|$")
    ax[0].grid(True)
    if label:
        ax[0].legend()

    # Plot phase φ
    ax[1].plot(
        freq, np.angle(Z, deg=deg), fmt, label=label, markersize=markersize, alpha=alpha
    )
    ax[1].set_xscale("log")
    ax[1].set_xlabel("frequency (Hz)")
    ax[1].set_ylabel(rf"$\phi$ ({'deg' if deg else 'rad'})")
    ax[1].grid(True)
    if label:
        ax[1].legend()

    if log:
        ax[0].set_yscale("log")
    ax[0].figure.get_layout_engine().set(wspace=0.1)

    return ax


def plot_impedance_combo(
    freq: np.ndarray,
    Z: np.ndarray,
    *,
    fmt: str = ".-",
    markersize: int = 6,
    ax: Iterable[plt.Axes] = None,
    label: str = None,
    log: bool = False,
) -> list[plt.Axes]:
    """Plots EIS data in Nyquist and Bode plots.

    Parameters
    ----------
    freq: np.ndarray[float]
        Frequencies corresponding to the impedance data.
    Z: np.ndarray[complex]
        Impedance data.
    fmt: str, optional
        Format of the markers in the plot. Default is ".-".
    markersize: int, optional
        Size of the markers in the plots. Default is 10.
    ax: Iterable[plt.Axes], optional
        Iterable of axes (must be of length 3) to plot on. Default is None.
    label: str, optional
        Label for the plot. Default is None.
    log: bool, optional
        If True, plots the y-axis of |Z| vs. freq in log scale. Default is False.

    Returns
    -------
    list[plt.Axes]
        List of axes objects of the Nyquist and Bode plots.
    """
    if ax is None:
        fig, ax = plt.subplots(ncols=3, figsize=(12, 3.25), layout="constrained")
    else:
        msg = "Incompatible 'ax'. Use fig, ax = plt.subplots(ncols=2), and pass 'ax'."
        assert len(ax) == 3 and all(isinstance(a, Axes) for a in ax), msg

    plot_nyquist(Z=Z, label=label, ax=ax[0], fmt=fmt, markersize=markersize)
    plot_bode(freq, Z, ax=ax[1:], fmt=fmt, markersize=markersize, log=log)
    ax2 = convert_to_twinx(ax[1:])

    return np.array([ax[0], *ax2])


def plot_linKK_residuals(
    freq: np.ndarray[float],
    res_real: np.ndarray[float],
    res_imag: np.ndarray[float],
    ax: plt.Axes = None,
) -> plt.Axes:
    """Plots the residuals of the linear Kramers-Kronig validation.

    Parameters
    ----------
    freq: np.ndarray[float]
        Frequencies corresponding to the residuals.
    res_real: np.ndarray[float]
        Real part of the residuals.
    res_imag: np.ndarray[float]
        Imaginary part of the residuals.
    ax: plt.Axes, optional
        Axes to plot on. Default is None.

    Returns
    -------
    plt.Axes
        Axes object of the residuals plot.
    """
    if ax is None:
        fig, ax = plt.subplots(figsize=(5.5, 3.5))
    ax.plot(freq, res_real, label="delta Re")
    ax.plot(freq, res_imag, label="delta Im")
    ax.set_xlabel("frequency (Hz)")
    ax.set_ylabel("delta %")
    ax.set_xscale("log")
    ax.set_title("Lin-KK validation")
    ax.ticklabel_format(axis="y", style="sci", scilimits=(-2, 2))
    ax.legend()
    return ax


def print_summary_statistics(mcmc: "numpyro.MCMC", circuit: str):
    """Prints summary statistics for the MCMC run."""
    # Set up required Rich objects
    console = Console()
    num_div = mcmc.get_extra_fields()["diverging"].sum()
    num_samples = mcmc.get_extra_fields()["diverging"].size
    title = f"{circuit}, {num_div}/{num_samples} divergences"
    table = Table(title=title, show_header=True, header_style="bold")

    # Add columns to the table
    columns = ["Parameter", "Mean", "Std", "Median", "5.0%", "95.0%"]
    for column in columns:
        table.add_column(column, justify="right")

    # Fill the table with data
    for label, values in mcmc.get_samples().items():
        rows = [
            label,
            f"{np.mean(values):.2e}",
            f"{np.std(values):.2e}",
            f"{np.median(values):.2e}",
            f"{np.percentile(values, 5):.2e}",
            f"{np.percentile(values, 95):.2e}",
        ]
        # Highlight rows with high standard deviation
        row_style = "on yellow" if np.std(values) > np.mean(values) else ""
        # Add a row to the table
        table.add_row(*rows, style=row_style)

    # Print the table
    console.print(table)


def print_inference_results(circuits: pd.DataFrame, return_table=True) -> Styler | Table:
    """Prints the inference results in a pretty format, excluding unncessary
    columns, highlighting the best performing circuits.

    Parameters
    ----------
    circuits : pd.DataFrame
        Circuits dataframe with inference results

    Returns
    -------
    pd.io.formats.style.Styler | Table
        Styled table with the inference results
    """
    circuits = circuits.copy(deep=True)

    # Rank the circuits based on WAIC
    if "WAIC (real)" in circuits.columns:
        circuits["WAIC (sum)"] = circuits["WAIC (real)"] + circuits["WAIC (imag)"]
    else:
        circuits["WAIC (sum)"] = circuits["WAIC (mag)"] + circuits["WAIC (phase)"]
    circuits.sort_values(by=["WAIC (sum)"], ascending=True, inplace=True, ignore_index=True)

    cols_to_hide = [
        "Parameters",
        "InferenceResult",
        "converged",
        "divergences",
        "Z_pred",
        "WAIC (sum)",
        "R^2 (real)",
        "R^2 (imag)",
        "MAPE (real)",
        "MAPE (imag)",
    ]
    df = circuits.style.hide(cols_to_hide, axis=1)
    fmt = {
        "WAIC (real)" if "WAIC (real)" in circuits.columns else "WAIC (mag)": "{:.2e}",
        "WAIC (imag)" if "WAIC (imag)" in circuits.columns else "WAIC (phase)": "{:.2e}",
        "R^2 (ravg)": "{:.3f}",
        "R^2 (iavg)": "{:.3f}",
        "MAPE (ravg)": "{:.2e}",
        "MAPE (iavg)": "{:.2e}",
    }
    df.format(fmt)

    # Create a rich Table to pretty print the results
    table = Table(title="Inference results", show_header=True, header_style="bold")

    # Add columns to the table
    columns = [
        "Circuit",
        "WAIC (real)" if "WAIC (real)" in circuits.columns else "WAIC (mag)",
        "WAIC (imag)" if "WAIC (imag)" in circuits.columns else "WAIC (phase)",
        "R2 (re)",
        "R2 (im)",
        "MAPE (re)",
        "MAPE (im)",
        "Np",
    ]
    for column in columns:
        table.add_column(column, justify="right")

    # Fill the table with data
    for i, row in df.data.iterrows():
        table.add_row(
            row["circuitstring"],
            f"{row['WAIC (real)' if 'WAIC (real)' in circuits.columns else 'WAIC (mag)']:.2e}",
            f"{row['WAIC (imag)' if 'WAIC (imag)' in circuits.columns else 'WAIC (phase)']:.2e}",
            f"{row['R^2 (ravg)']:.3f}",
            f"{row['R^2 (iavg)']:.3f}",
            f"{row['MAPE (ravg)']:.2e}",
            f"{row['MAPE (iavg)']:.2e}",
            f"{row['n_params']}",
        )

    return table if return_table else df


def override_mpl_colors(override_named_colors: bool = True):
    """Override matplotlib's default colors with Flexoki colors."""
    # Define the Flexoki-Light color scheme based on the provided table
    # Original sequence: red, orange, yellow, green, cyan, blue, purple, magenta
    flexoki_light_colors = {
        "red": mcolors.to_rgb("#D14D41"),
        "blue": mcolors.to_rgb("#4385BE"),
        "green": mcolors.to_rgb("#879A39"),
        "orange": mcolors.to_rgb("#DA702C"),
        "purple": mcolors.to_rgb("#8B7EC8"),
        "yellow": mcolors.to_rgb("#D0A215"),
        "cyan": mcolors.to_rgb("#3AA99F"),
        "magenta": mcolors.to_rgb("#CE5D97"),
    }

    # Override default named colors
    if override_named_colors:
        for k, v in flexoki_light_colors.items():
            mcolors.ColorConverter.cache[(k, None)] = v
            mcolors.ColorConverter.colors[k] = v
            mcolors.ColorConverter.cache[(k[0], None)] = v  # k[0] is the short name
            mcolors.ColorConverter.colors[k[0]] = v

    # Define the Flexoki-Light style
    flexoki_light_style = {
        "axes.prop_cycle": mpl.cycler(
            color=list(flexoki_light_colors.values()),
            linestyle=["-", "--", "-.", ":"] * 2,
        ),
        "axes.facecolor": "white",
        "axes.edgecolor": "black",
        "axes.grid": True,
        "axes.axisbelow": True,
        "axes.labelcolor": "black",
        "figure.facecolor": "white",
        "grid.color": "whitesmoke",
        "text.color": "black",
        "xtick.color": "black",
        "ytick.color": "black",
        "xtick.direction": "out",
        "ytick.direction": "out",
        "lines.color": flexoki_light_colors["blue"],
        "patch.edgecolor": "black",
        # "axes.spines.top": False,
        # "axes.spines.right": False
    }

    # Apply the Flexoki-Light style
    plt.style.use(flexoki_light_style)


def set_plot_style(
    use_arviz: bool = True,
    use_seaborn: bool = True,
    use_flexoki: bool = True,
    console_width: int = 100,
):
    """Modifies the default arviz/matplotlib config for prettier plots.

    Parameters
    ----------
    use_arviz: bool, optional
        If True, use arviz's default plotting style. Default is True.
    use_seaborn: bool, optional
        If True, use seaborn's default plotting style. Default is True.
    use_flexoki: bool, optional
        If True, use Flexoki's default plotting style. Default is True.
    console_width: int, optional
        Width of the console in characters. Only set when in VS Code
        Interactive mode, otherwise automaticall deterimined. Default is 100.
    """
    # Arviz
    if use_arviz:
        arviz.style.use("arviz-viridish")

    # Seaborn
    if use_seaborn:
        sns.set_style("ticks")

    # Matplotlib
    label_size = 11
    tick_size = label_size - 1
    ax_title_size = label_size + 1
    fig_title_size = ax_title_size + 2
    legend_size = label_size - 1

    plt.rcParams["font.family"] = "sans-serif"
    fonts = ["Helvetica", "Arial", "Verdana", "Tahoma", "DejaVu Sans"]
    plt.rcParams["font.sans-serif"] = fonts
    plt.rcParams["mathtext.fontset"] = "dejavuserif"
    plt.rcParams["xtick.labelsize"] = tick_size
    plt.rcParams["ytick.labelsize"] = tick_size
    plt.rcParams["axes.labelsize"] = label_size
    plt.rcParams["axes.titlesize"] = ax_title_size
    plt.rcParams["figure.titlesize"] = fig_title_size
    plt.rcParams["legend.fontsize"] = legend_size
    plt.rcParams["legend.frameon"] = False

    # Flexoki colors
    if use_flexoki:
        override_mpl_colors()

    # Customize interactive sessions
    try:
        if is_interactive_session():
            import matplotlib_inline.backend_inline

            matplotlib_inline.backend_inline.set_matplotlib_formats("retina")
    except ImportError:
        pass

    # Set rich console width when in VS Code Interactive Window
    if os.getenv("VSCODE_PID") is not None:
        rich.get_console().width = console_width


def show_nticks(ax: plt.Axes, x: bool = True, y: bool = False, n: int = 10):
    """In-place modifies Matplotlib axes to show only ``n`` ticks.

    Parameters
    ----------
    ax: plt.Axes
        Axes to modify.
    x: bool, optional
        If True, applies the filter to the x-axis. Default is True.
    y: bool, optional
        If True, applies the filter to the y-axis. Default is False.
    n: int, optional
        Number of ticks to show. Default is 10.
    """
    if x:
        xticks = ax.xaxis.get_major_ticks()
        if len(xticks) > n:
            ax.xaxis.set_major_locator(plt.MaxNLocator(n, steps=[1, 2, 5, 10]))
    if y:
        yticks = ax.yaxis.get_major_ticks()
        if len(yticks) > n:
            ax.yaxis.set_major_locator(plt.MaxNLocator(n, steps=[1, 2, 5, 10]))


def convert_to_twinx(ax: np.ndarray[plt.Axes]) -> tuple[plt.Axes]:
    """Convert a 2-column subplots ax object into a single-column with twinx.
    Left axis is always red, right axis is always blue.

    Parameters
    ----------
    ax : numpy.ndarray
        Array of 2-column subplot Axes objects.

    Returns
    -------
    tuple[plt.Axes]
        Two Axes objects with the same data from the original subplots, but
        now combined into a single-column plot with twinx.
    """
    # Check that the input is a 2-column axes object
    if ax.size != 2:
        raise ValueError("Input 'ax' object must have 2 subplots.")

    # Extract the first subplot as the base for the new single-column plot
    ax1 = ax[0]

    # Get the data and labels from the second column (right plot)
    ax2 = ax1.twinx()  # Create a twin axis sharing the x-axis

    # Copy over the data from the second subplot to the right axis
    for line in ax[1].get_lines():
        ax2.plot(
            line.get_xdata(),
            line.get_ydata(),
            color="b",  # Set twin axis plot to blue
            label=line.get_label(),
            linestyle=line.get_linestyle(),
            marker=line.get_marker(),
        )

    # Set labels and titles from the original plots
    ax1.set_xlabel(ax[0].get_xlabel())
    ax1.set_ylabel(ax[0].get_ylabel(), color="r")  # Set left axis label to red
    ax2.set_ylabel(ax[1].get_ylabel(), color="b")  # Set right axis label to blue

    # Remove the right subplot to clean up the figure
    ax[1].remove()

    # Turn off grid for the 2nd axis
    ax2.grid(False)

    # Set tick parameters to match the axis colors
    ax1.tick_params(axis="y", colors="r")  # Left axis red ticks
    ax2.tick_params(axis="y", colors="b")  # Right axis blue ticks

    # Ensure the lines match the designated axis colors
    for l1 in ax[0].get_lines():
        l1.set_color("r")  # Set left axis lines to red
    for l2 in ax2.get_lines():
        l2.set_color("b")  # Set right axis lines to blue

    return ax1, ax2



"""
Collection of functions for parsing circuit strings.

.. currentmodule:: autoeis.parser

.. autosummary::
   :toctree: generated/

    validate_circuit
    validate_parameter
    parse_component
    parse_parameter
    get_component_labels
    get_component_types
    get_parameter_types
    get_parameter_labels
    group_parameters_by_type
    group_parameters_by_component
    count_parameters
    convert_to_impedance_format
    find_ohmic_resistors
    simplify

"""

import re
from functools import lru_cache

from numpy import pi  # noqa: F401
from pyparsing import nested_expr


def validate_circuit(circuit: str) -> bool:
    """Checks if a circuit string is valid.

    This function ensures that the circuit string:
        - is not empty,
        - contains valid element names (``R``, ``C``, ``L``, ``P``),
        - and, doesn't contain duplicate elements, e.g., ``R1-[P2,P2]``.

    Parameters
    ----------
    circuit : str
        CDC string representation of the input circuit. See
        `here <https://autodial.github.io/AutoEIS/circuit.html>`_ for details.

    Returns
    -------
    bool
        True if the circuit string is valid, False otherwise.
    """
    # TODO: Check for parallel elements with < 2 elements
    # TODO: Check for duplicate "-" or "," symbols
    # TODO: Check for disconnected elements, eg R1R2 or R1[R2,R3]
    # Check for duplicate elements
    components = get_component_labels(circuit)
    duplicates = [e for e in components if components.count(e) > 1]
    assert not duplicates, f"Duplicate elements found: {set(duplicates)}"
    # Test circuit is not empty
    assert len(circuit) > 0, "Circuit string is empty."
    # Check for valid element names
    valid_types = ["R", "C", "L", "P"]
    types = get_component_types(circuit)
    for t in types:
        assert t in valid_types, f"Invalid element type: {t}"
    return True  # If all checks pass, the circuit is considered valid


def validate_parameter(p: str, raises: bool = True) -> bool:
    """Checks if a parameter label is valid.

    Valid parameter labels: {R,C,L,Pw,Pn}{N} where N is a number, e.g., P1n.

    Parameters
    ----------
    p : str
        String representation of the parameter label.
    raises : bool, optional
        If True, raises an AssertionError on invalid parameter labels.
        Default is True.

    Returns
    -------
    bool
        True if the parameter label is valid, False otherwise.
    """
    # Check if parameter label is a string
    if not isinstance(p, str):
        if raises:
            raise AssertionError("Parameter label must be a string.")
        return False
    # Check if parameter label is not empty
    if not p:
        if raises:
            raise AssertionError("Parameter label is empty.")
        return False
    # Check if parameter label is valid
    pattern = r"(?:R\d+|C\d+|L\d+|P\d+[wn])"
    if not re.fullmatch(pattern, p):
        if raises:
            raise AssertionError(f"Invalid parameter label: {p}")
        return False
    return True


def parse_component(c: str) -> str:
    """Returns the component type of a component/parameter label.

    Parameters
    ----------
    c : str
        String representation of a component/parameter label.

    Returns
    -------
    str
        The type of the component label from the set {R,C,L,P}.

    Examples
    --------
    >>> parse_component("R1")
    'R'
    >>> parse_component("P2n")
    'P'
    """
    return re.match(r"[A-Za-z]+", c).group()


def parse_parameter(p: str) -> str:
    """Returns the type of a parameter label.

    Parameters
    ----------
    p : str
        String representation of the parameter label.

    Returns
    -------
    str
        The type of the parameter label from the set {R,C,L,Pn,Pw}.

    Examples
    --------
    >>> parse_parameter("R1")
    'R'
    >>> parse_parameter("P2n")
    'Pn'
    """
    validate_parameter(p)
    if p.startswith(("R", "C", "L")):
        ptype = p[0]
    elif p.startswith("P") and p.endswith("w"):
        ptype = "Pw"
    elif p.startswith("P") and p.endswith("n"):
        ptype = "Pn"
    return ptype


def get_component_labels(circuit: str, types: list[str] = None) -> list[str]:
    """Returns a list of labels for all components in a circuit string.

    Parameters
    ----------
    circuit : str
        CDC string representation of the input circuit. See
        `here <https://autodial.github.io/AutoEIS/circuit.html>`_ for details.
    types : list[str], optional
        List of component types to filter by. Default is None.

    Returns
    -------
    list[str]
        A list of component labels.

    Examples
    --------
    >>> get_component_labels("R1-[R2,P4]")
    ['R1', 'R2', 'P4']
    >>> get_component_labels("R1-[R2,P4]", types=["R"])
    ['R1', 'R2']
    """
    types = [types] if isinstance(types, str) else types
    types = ["R", "C", "L", "P"] if types is None else types
    assert isinstance(types, list), "types must be a list of strings."
    pattern = rf"\b(?:{'|'.join(types)})\d+\b"
    return re.findall(pattern, circuit)


def get_component_types(circuit: str, unique: bool = False) -> list[str]:
    """Returns a list of component types in a circuit string.

    Parameters
    ----------
    circuit : str
        CDC string representation of the input circuit. See
        `here <https://autodial.github.io/AutoEIS/circuit.html>`_ for details.
    unique : bool, optional
        If True, returns a list of unique component types. Default is False.

    Returns
    -------
    list[str]
        A list of component types.

    Examples
    --------
    >>> get_component_types("R1-[R2,P4]")
    ['R', 'R', 'P']
    >>> get_component_types("R1-[R2,P4]", unique=True)
    ['P', 'R']
    """
    types = re.findall(r"[A-Za-z]+", circuit)
    return list(set(types)) if unique else types


def get_parameter_labels(circuit: str, types: list[str] = None) -> list[str]:
    """Returns a list of labels for all parameters in a circuit string.

    Parameters
    ----------
    circuit : str
        CDC string representation of the input circuit. See
        `here <https://autodial.github.io/AutoEIS/circuit.html>`_ for details.
    types : list[str], optional
        List of parameter types to filter by. Default is None.

    Returns
    -------
    list[str]
        A list of parameter labels.

    Examples
    --------
    >>> get_parameter_labels("R1-[R2,P4]")
    ['R1', 'R2', 'P4w', 'P4n']
    >>> get_parameter_labels("R1-[R2,P4]", types=["R"])
    ['R1', 'R2']
    """
    types = [types] if isinstance(types, str) else types
    types = ["R", "C", "L", "P"] if types is None else types
    assert isinstance(types, list), "types must be a list of strings."
    components = get_component_labels(circuit, types=types)
    parameters = []
    for component in components:
        # CPE elements have two parameters P{i}w and P{i}n
        if component.startswith("P"):
            parameters.extend([f"{component}w", f"{component}n"])
        else:
            parameters.append(component)
    return parameters


def get_parameter_types(circuit: str, unique: bool = False) -> list[str]:
    """Returns a list of parameter types in a circuit string.

    Parameters
    ----------
    circuit : str
        CDC string representation of the input circuit. See
        `here <https://autodial.github.io/AutoEIS/circuit.html>`_ for details.
    unique : bool, optional
        If True, returns a list of unique parameter types. Default is False.

    Returns
    -------
    list[str]
        A list of parameter types.

    Examples
    --------
    >>> get_parameter_types("R1-[R2,P4]")
    ['R', 'R', 'Pw', 'Pn']
    >>> get_parameter_types("R1-[R2,P4]", unique=True)
    ['Pn', 'Pw', 'R']
    """
    ptypes = [parse_parameter(p) for p in get_parameter_labels(circuit)]
    return list(set(ptypes)) if unique else ptypes


def group_parameters_by_type(circuit: str) -> dict[str, list[str]]:
    """Groups parameter labels by component type.

    Parameters
    ----------
    circuit : str
        CDC string representation of the input circuit. See
        `here <https://autodial.github.io/AutoEIS/circuit.html>`_ for details.

    Returns
    -------
    dict[str, list[str]]
        A dictionary of parameter labels grouped by component type.

    Examples
    --------
    >>> group_parameters_by_type("R1-[R2,P4]")
    {'Pn': ['P4n'], 'Pw': ['P4w'], 'R': ['R1', 'R2']}
    """
    params = get_parameter_labels(circuit)
    ptypes = get_parameter_types(circuit)
    groups = {ptype: [] for ptype in set(ptypes)}
    for param, ptype in zip(params, ptypes):
        groups[ptype].append(param)
    return groups


def group_parameters_by_component(circuit: str) -> dict[str, list[str]]:
    """Groups parameter labels by component label.

    Parameters
    ----------
    circuit : str
        CDC string representation of the input circuit. See
        `here <https://autodial.github.io/AutoEIS/circuit.html>`_ for details.

    Returns
    -------
    dict[str, list[str]]
        A dictionary of parameter labels grouped by component label.

    Examples
    --------
    >>> group_parameters_by_component("R1-[R2,P4]")
    {'R': ['R1', 'R2'], 'P': ['P4w', 'P4n']}
    """
    ctypes = get_component_types(circuit)
    params_by_component = {ctype: [] for ctype in ctypes}
    params = get_parameter_labels(circuit)
    for param in params:
        ctype = parse_component(param)
        params_by_component[ctype].append(param)
    return params_by_component


def count_parameters(circuit: str) -> int:
    """Returns the number of parameters present in a circuit string.

    Parameters
    ----------
    circuit : str
        CDC string representation of the input circuit. See
        `here <https://autodial.github.io/AutoEIS/circuit.html>`_ for details.

    Returns
    -------
    int
        The number of parameters present in the circuit.

    Examples
    --------
    >>> count_parameters("R1-[R2,P4]")
    4
    """
    return len(get_parameter_labels(circuit))


def convert_to_impedance_format(circuit: str) -> str:
    """Converts a circuit string the format used by impedance.py.

    Parameters
    ----------
    circuit : str
        CDC string representation of the input circuit. See
        `here <https://autodial.github.io/AutoEIS/circuit.html>`_ for details.

    Returns
    -------
    str
        The circuit string in the format used by impedance.py.

    Examples
    --------
    >>> convert_to_impedance_format("R1-[R2,P4]")
    'R1-p(R2,CPE4)'
    """
    circuit = circuit.replace("P", "CPE")
    circuit = circuit.replace("[", "p(")
    circuit = circuit.replace("]", ")")
    return circuit


def circuit_to_nested_expr(circuit: str) -> list:
    """Parses a circuit string to a nested list[str].

    Parameters
    ----------
    circuit : str
        CDC string representation of the input circuit. See
        `here <https://autodial.github.io/AutoEIS/circuit.html>`_ for details.

    Returns
    -------
    list
        A nested list of component labels.

    Examples
    --------
    >>> circuit_to_nested_expr("R1-[R2,P4]")
    ['R1', ['R2,P4']]
    """

    def cleanup(lst: list, chars: list[str]):
        """Removes leading/trailing chars from a nested list[str]."""
        chars = "".join(chars)
        result = []
        for el in lst:
            if isinstance(el, list):
                result.append(cleanup(el, chars))
            else:
                # Don't add empty strings
                if el.strip(chars):
                    result.append(el.strip(chars))
        return result

    def parse(circuit: str):
        # Enclose circuit with brackets to make it a valid nested expression
        circuit = f"[{circuit}]"
        parser = nested_expr(opener="[", closer="]")
        parsed = parser.parse_string(circuit, parse_all=True).as_list()
        return parsed

    expr = parse(circuit)
    expr = cleanup(expr, chars=[",", "-"])
    return expr[0]


def find_series_elements(circuit: str) -> list[str]:
    """Extracts the series componenets from a circuit (in the main chain).

    Parameters
    ----------
    circuit : str
        CDC string representation of the input circuit. See
        `here <https://autodial.github.io/AutoEIS/circuit.html>`_ for details.

    Returns
    -------
    list[str]
        A list of series component labels.

    Examples
    --------
    >>> find_series_elements("R1-[R2,P4]-P5")
    ['R1', 'P5]
    """
    parsed = circuit_to_nested_expr(circuit)
    series_elements = [el for el in parsed if isinstance(el, str)]
    series_elements = re.findall(r"[A-Z]+\d+", str(series_elements))
    return series_elements


def find_ohmic_resistors(circuit: str) -> list[str]:
    """Finds all ohmic resistors in a circuit (only in the main chain).

    Parameters
    ----------
    circuit : str
        CDC string representation of the input circuit. See
        `here <https://autodial.github.io/AutoEIS/circuit.html>`_ for details.

    Returns
    -------
    list[str]
        A list of ohmic resistor labels.

    Examples
    --------
    >>> find_ohmic_resistors("R1-[R2,P4]-R5")
    ['R1', 'R5']
    """
    series_elements = find_series_elements(circuit)
    return re.findall(r"R\d+", str(series_elements))


def generate_mathematical_expression(circuit: str) -> str:
    """Converts a circuit string to a mathematical expression, parameterized
    by frequency and the circuit parameters, i.e., func(freq, p).

    The returned string can be evaluated assuming 'p' is an array of
    parameter values and 'freq' is the frequency (scalar/array).

    Parameters
    ----------
    circuit : str
        CDC string representation of the input circuit. See
        `here <https://autodial.github.io/AutoEIS/circuit.html>`_ for details.

    Returns
    -------
    str
        The mathematical expression for impedance.

    Examples
    --------
    >>> generate_mathematical_expr("R1-R2")
    'p[0]+(1/(p[1]*(2*1j*pi*freq)**p[2]))'
    """
    # Apply series-parallel conversion, e.g., [R1,R2] -> (1/R1+1/R2)**(-1)
    replacements = {
        "-": "+",
        "[": "((",
        ",": ")**(-1)+(",
        "]": ")**(-1))**(-1)"
    }  # fmt: off
    expr = circuit
    for j, k in replacements.items():
        expr = expr.replace(j, k)

    # Embed impedance expressions, e.g., C1 -> (1/(2*1j*pi*f*C1))
    expr = replace_components_with_impedance(expr)

    # Replace parameters with array indexing, e.g., R1, P2w, P2n -> x[0], x[1], x[2]
    parameters = get_parameter_labels(circuit)
    for i, var in enumerate(parameters):
        # Negative look-ahead to avoid replacing R10 when dealing with R1
        expr = re.sub(rf"{var}(?!\d)", f"p[{i}]", expr)

    return expr


def replace_components_with_impedance(expr: str) -> str:
    """Expands the circuit expression with the impedance of the components.

    The circuit expression describes describes the impedance of a circuit,
    parameterized by component impedance values, e.g., '[R1,P2]' ->
    '1 / (1/R1 + 1/P2)'. This function parameterizes the impedence terms using
    the actual component values, e.g., R1 -> R1, P2 -> 1/(P2w*(2*1j*pi*freq)**P2n)

    Parameters
    ----------
    expr : str
        The circuit expression to be expanded.

    Returns
    -------
    str
        The expanded circuit expression.

    Examples
    --------
    >>> replace_components_with_impedance("R1")
    'R1'
    >>> replace_components_with_impedance("P1")
    '(1/(P1w*(2*1j*pi*freq)**P1n))'
    >>> replace_components_with_impedance("R1+P2")
    'R1+(1/(P2w*(2*1j*pi*freq)**P2n))'
    """

    def replacement(var):
        eltype = get_component_types(var)[0]
        return {
            "R": f"{var}",
            "C": f"(1/(2*1j*pi*freq*{var}))",
            "L": f"(2*1j*pi*freq*{var})",
            "P": f"(1/({var}w*(2*1j*pi*freq)**{var}n))",
        }[eltype]

    # Get component lables
    components = get_component_labels(expr)
    # Replace components with impedance expression, e.g., C1 -> (1/(2*1j*pi*freq*C1))
    for c in components:
        # Negative look-ahead to avoid replacing R10 when dealing with R1
        expr = re.sub(rf"{c}(?!\d)", replacement(c), expr)

    return expr


def _find_top_level_components(s: str, delimiter: str) -> list[str]:
    """
    Splits a string by a delimiter, but only when the delimiter is not
    inside brackets. This is a robust way to find top-level components.
    """
    components = []
    bracket_level = 0
    start = 0
    for i, char in enumerate(s):
        if char == "[":
            bracket_level += 1
        elif char == "]":
            bracket_level -= 1
        elif char == delimiter and bracket_level == 0:
            components.append(s[start:i])
            start = i + 1
    components.append(s[start:])
    return components


def _parse_to_structure(s: str):
    """
    Recursively parses the circuit string into a nested list structure (an AST).
    'R1-[R2,C1]' -> ['s', 'R1', ['p', 'R2', 'C1']]
    """
    # Base case: The component is just an element.
    if "-" not in s and "[" not in s:
        return s

    # It's a series of components.
    series_components = _find_top_level_components(s, "-")

    # If there's only one component after splitting, it might be a parallel block.
    if len(series_components) == 1:
        component = series_components[0]
        if component.startswith("[") and component.endswith("]"):
            content = component[1:-1]
            branches = _find_top_level_components(content, ",")
            # If there's only one branch, it's not a true parallel block.
            if len(branches) == 1:
                return _parse_to_structure(branches[0])
            return ["p"] + [_parse_to_structure(branch) for branch in branches]
        return component

    return ["s"] + [_parse_to_structure(comp) for comp in series_components]


def _simplify_structure(struct):
    """
    Recursively simplifies the nested list structure in a single pass.
    """
    # Base case: An element (string) cannot be simplified further.
    if isinstance(struct, str):
        return struct

    op = struct[0]
    components = struct[1:]

    # Recursively simplify children first (post-order traversal).
    simplified_components = [_simplify_structure(c) for c in components]

    # --- Apply Simplification Logic ---
    final_components = []
    seen_types = set()
    simplifiable_types = {"R", "C", "L"}

    if op == "s":  # Series simplification
        for comp in simplified_components:
            comp_type = ""
            if isinstance(comp, str):
                comp_type = comp[0]
            elif comp[0] == "p":  # It's a parallel block
                comp_type = "p"

            if comp_type in simplifiable_types:
                if comp_type not in seen_types:
                    seen_types.add(comp_type)
                    final_components.append(comp)
            else:  # P elements or parallel blocks
                final_components.append(comp)

    elif op == "p":  # Parallel simplification
        for comp in simplified_components:
            is_atomic = isinstance(comp, str) and comp[0] != "p"

            if is_atomic:
                comp_type = comp[0]
                if comp_type in simplifiable_types:
                    if comp_type not in seen_types:
                        seen_types.add(comp_type)
                        final_components.append(comp)
                else:  # P element
                    final_components.append(comp)
            else:  # Complex branch
                final_components.append(comp)

    # If only one component remains after simplification, unwrap it.
    if len(final_components) == 1:
        return final_components[0]

    return [op] + final_components


def _stringify_structure(struct) -> str:
    """
    Recursively builds the final string from the simplified structure.
    """
    if isinstance(struct, str):
        return struct

    op = struct[0]
    stringified_components = [_stringify_structure(c) for c in struct[1:]]

    if op == "s":
        return "-".join(stringified_components)
    # op == 'p'
    return f"[{','.join(stringified_components)}]"


@lru_cache(maxsize=1_000_000)
def simplify(circuit: str) -> str:
    """
    Simplifies a circuit string to its canonical form by applying series-parallel
    reduction, e.g., 'R1-[C2,C3]' -> 'R1-C2' or 'R1-R2' -> 'R1'.

    Parameters
    ----------
    circuit : str
        CDC string representation of the input circuit. See
        `here <https://autodial.github.io/AutoEIS/circuit.html>` for details.

    Returns
    -------
    str
        The simplified circuit string in canonical form.
    """
    if not circuit:
        return ""

    structure = _parse_to_structure(circuit)
    simplified_structure = _simplify_structure(structure)
    return _stringify_structure(simplified_structure)



"""
Collection of functions for importing and exporting EIS data/results.

.. currentmodule:: autoeis.io

.. autosummary::
   :toctree: generated/

    get_assets_path
    load_battery_dataset
    load_test_dataset
    load_test_circuits
    parse_ec_output

"""

import logging
import os
from collections.abc import Iterable
from pathlib import Path

import numpy as np
import pandas as pd

import autoeis as ae

log = logging.getLogger(__name__)


def get_assets_path() -> Path:
    """Returns the path to the assets folder."""
    PATH = Path(ae.__file__).parent / "assets"
    return PATH


def load_battery_dataset(
    preprocess: bool = False,
    noise: float = 0,
) -> list[tuple[np.ndarray[float], np.ndarray[complex]]]:
    """Loads EIS data of a battery cell during cycling (at discharged state).

    Parameters
    ----------
    preprocess: bool, optional
        If True, the impedance data is preprocessed using
        :func:`autoeis.core.preprocess_impedance_data`. Default is False.
    noise: float, optional
        If greater than zero, uniform noise with prescribed amplitude is added
        to the impedance data. Default is 0.

    Returns
    -------
    list[tuple[np.ndarray[float], np.ndarray[complex]]]
        List of tuples of frequency and impedance arrays for each cycle.
    """
    PATH = get_assets_path()
    fpath = os.path.join(PATH, "battery_data.npy")
    data = np.load(fpath)
    # Data are stored as complex, convert frequency to float
    data = [(freq.real, Z) for freq, Z in data]
    if preprocess:
        data = [ae.utils.preprocess_impedance_data(freq, Z) for freq, Z in data]
    if noise:
        # Only add noise to impedance data (opinionated!)
        noise_real = [np.random.rand(len(Z)) * Z.real * noise for freq, Z in data]
        noise_imag = [np.random.rand(len(Z)) * Z.imag * noise for freq, Z in data]
        data = [
            (freq, Z + noise_real[i] + noise_imag[i] * 1j) for i, (freq, Z) in enumerate(data)
        ]
    return data


def load_test_dataset(
    preprocess: bool = False, noise: float = 0
) -> tuple[np.ndarray[float], np.ndarray[complex]]:
    """Returns a test dataset as a tuple of frequency and impedance arrays.

    Parameters
    ----------
    preprocess: bool, optional
        If True, the impedance data is preprocessed using
        :func:`autoeis.core.preprocess_impedance_data`. Default is False.
    noise: float, optional
        If greater than zero, uniform noise with prescribed amplitude is added
        to the impedance data. Default is 0.

    Returns
    -------
    tuple[np.ndarray[float], np.ndarray[complex]]
        Tuple of frequency and impedance data.
    """
    PATH = get_assets_path()
    fpath = os.path.join(PATH, "test_data.txt")
    freq, Zreal, Zimag = np.loadtxt(fpath, skiprows=1, unpack=True, usecols=(0, 1, 2))
    # Convert to complex impedance (the file contains -Im(Z) hence the minus sign)
    Z = Zreal - 1j * Zimag
    if preprocess:
        freq, Z = ae.utils.preprocess_impedance_data(freq, Z)
    if noise:
        # Only add noise to impedance data (opinionated!)
        noise_real = np.random.rand(len(Z)) * Z.real * noise
        noise_imag = np.random.rand(len(Z)) * Z.imag * noise
        Z += noise_real + noise_imag * 1j
    return freq, Z


def load_test_circuits(filtered: bool = False) -> pd.DataFrame:
    """Returns candidate ECMs fitted to test dataset for testing.

    Parameters
    ----------
    filtered: bool, optional
        If True, only physically plausible circuits are returned. Default is False.

    Returns
    -------
    pd.DataFrame
        Dataframe containing the ECMs.

    """
    PATH = get_assets_path()
    fname = "circuits_filtered.csv" if filtered else "circuits_unfiltered.csv"
    fpath = os.path.join(PATH, fname)
    circuits = pd.read_csv(fpath)
    # Convert stringified list to proper Python objects
    circuits["Parameters"] = circuits["Parameters"].apply(eval)
    return circuits


def parse_ec_output(
    circuits: Iterable[str] | str, ignore_invalid_inputs: bool = True
) -> pd.DataFrame:
    """Parses the output of EquivalentCircuits.jl's ``circuit_evolution``.

    Parameters
    ----------
    circuits: Iterable[str] | str
        List of stringified output of EquivalentCircuits.jl's
        ``circuit_evolution``. A valid input should be in the following format:
        'EquivalentCircuit("R1", (R1 = 1.0,))', or a list of such strings.

    Returns
    -------
    pd.DataFrame
        Dataframe containing ECMs (cols: "circuitstring" and "Parameters")
    """

    def _validate_input(ec_output: str, raise_error: bool = True):
        """Ensures the input string can be parsed into circuit and parameters."""
        ec_output = ec_output.replace(" ", "")
        try:
            assert len(ec_output.split('",(')) == 2
            assert "EquivalentCircuit" in ec_output
        except AssertionError:
            if raise_error:
                raise ValueError(f"Invalid EC output format: {ec_output}.")
            return False
        return True

    def _split_labels_and_values(ec_output: str) -> tuple[str, dict[str, float]]:
        """Splits the circuit string and parameters from the input string."""
        ec_output = ec_output.removeprefix("EquivalentCircuit(").removesuffix(")")
        ec_output = ec_output.replace(" ", "")
        cstr, pstr = ec_output.split('",(')
        cstr = cstr.replace('"', "")
        # NOTE: Trailing comma in one-element tuples needs to also be removed
        pstr = pstr.replace(")", "").replace("(", "").replace('"', "").rstrip(",").split(",")
        # Convert parameters substring to a dict[str, float]
        pdict = dict(pair.split("=") for pair in pstr)  # dict[str, str]
        pdict = {p.split("=")[0]: float(p.split("=")[1]) for p in pstr}  # dict[str, float]
        return cstr, pdict

    circuits = [circuits] if isinstance(circuits, str) else circuits
    parsed = []

    header = "circuitstring,Parameters"

    for circuit in circuits:
        # Validate input format
        circuit = circuit.replace(" ", "")
        # Skip header if present
        if circuit == header:
            continue
        # Skip invalid inputs if flag is set
        if not _validate_input(circuit, raise_error=not ignore_invalid_inputs):
            continue
        # Finally, parse the circuit and parameters
        cstr, pdict = _split_labels_and_values(circuit)
        parsed.append([cstr, pdict])

    return pd.DataFrame(parsed, columns=["circuitstring", "Parameters"])



import os
import logging
import shutil
from pathlib import Path

import juliapkg
from juliapkg.deps import can_skip_resolve
from juliapkg.find_julia import find_julia

from .utils import suppress_output
from .version import __equivalent_circuits_jl_version__

log = logging.getLogger(__name__)


def install_julia(quiet=True):
    """Installs Julia using juliapkg."""
    # Importing juliacall automatically installs Julia using juliapkg
    if quiet:
        with suppress_output():
            import juliacall
    else:
        import juliacall


def install_backend(ec_path: Path = None, quiet=True):
    """Installs Julia dependencies for AutoEIS.

    Parameters
    ----------
    ec_path : Path, optional
        Path to the local copy of EquivalentCircuits. Default is None. If None,
        the remote version will be used.
    """
    is_julia_installed(error=True)

    # TODO: No longer needed since dependencies are specified in juliapkg.json
    # kwargs = {"name": "EquivalentCircuits", "uuid": "da5bd070-f609-4e16-a30d-de86b3faa756"}
    # if ec_path is not None:
    #     kwargs["path"] = str(ec_path)
    #     kwargs["dev"] = True
    # else:
    #     if __equivalent_circuits_jl_version__.startswith("v"):
    #         kwargs["version"] = __equivalent_circuits_jl_version__
    #     else:
    #         kwargs["rev"] = __equivalent_circuits_jl_version__
    #         kwargs["url"] = "https://github.com/ma-sadeghi/EquivalentCircuits.jl"
    # pkg_spec = juliapkg.PkgSpec(**kwargs)
    # juliapkg.add(pkg_spec)

    if quiet:
        with suppress_output():
            juliapkg.resolve()
    else:
        juliapkg.resolve()


def init_julia(quiet=False):
    """Initializes Julia and returns the Main module.

    Raises
    ------
    ImportError
        If Julia is not installed.
    """
    is_julia_installed(error=True)
    if not can_skip_resolve():
        log.warning("Julia is installed, but needs to be resolved...")
    if quiet:
        with suppress_output():
            from juliacall import Main
    else:
        from juliacall import Main

    return Main


def import_package(package_name, Main, error=False):
    """Imports a package in Julia and returns the module.

    Parameters
    ----------
    package_name : str
        Name of the Julia package to import.
    Main : juliacall.Main
        Julia Main module.
    error : bool, optional
        If True, raises an error if the package is not found. Default is False.

    Returns
    -------
    module
        The imported Julia module.

    Raises
    ------
    ImportError
        If the package is not found and error is True.
    """
    from juliacall import JuliaError

    try:
        Main.seval(f"using {package_name}")
        return eval(f"Main.{package_name}")
    except JuliaError as e:
        if error:
            raise e
    return None


def import_backend(Main=None):
    """Imports EquivalentCircuits package from Julia.

    Parameters
    ----------
    Main : juliacall.Main, optional
        Julia Main module. Default is None.

    Returns
    -------
    module
        The imported Julia module.

    Raises
    ------
    ImportError
        If Julia is not installed or the package is not found.
    """
    Main = init_julia() if Main is None else Main
    is_backend_installed(Main=Main, error=True)
    return import_package("EquivalentCircuits", Main)


def is_julia_installed(error=False):
    """Asserts that Julia is installed."""
    # Look for system-wide Julia executable
    try:
        find_julia()
        return True
    except Exception:
        pass
    # Look for local Julia executable (e.g., installed by juliapkg)
    if can_skip_resolve():
        return True
    msg = "Julia not found. Visit https://github.com/JuliaLang/juliaup and install Julia."
    if error:
        raise ImportError(msg)
    return False


def is_backend_installed(Main=None, error=False):
    """Asserts that EquivalentCircuits.jl is installed.

    Parameters
    ----------
    Main : juliacall.Main, optional
        Julia Main module. Default is None. If None, the Main module will be
        initialized using `init_julia()`.
    error : bool, optional
        If True, raises an error if the package is not found. Default is False.
    install : bool, optional
        If True, installs the package if it is not found. Default is False.

    Returns
    -------
    bool
        True if the package is installed, False otherwise.

    Raises
    ------
    ImportError
        If Julia is not installed or the package is not found and error is True.
    """
    Main = init_julia() if Main is None else Main
    if import_package("EquivalentCircuits", Main, error=False) is not None:
        return True
    msg = "EquivalentCircuits.jl not found, run 'python -m autoeis install'"
    if error:
        raise ImportError(msg)
    return False


def ensure_julia_deps_ready(quiet=True, retry=True):
    """Ensures Julia and EquivalentCircuits.jl are installed."""

    def _ensure_julia_deps_ready(quiet):
        if not is_julia_installed(error=False):
            log.warning("Julia not found, installing Julia...")
            install_julia(quiet=quiet)
        Main = init_julia(quiet=quiet)
        if not is_backend_installed(Main=Main, error=False):
            log.warning("Julia dependencies not found, installing EquivalentCircuits.jl...")
            install_backend(quiet=quiet)

    def _reset_julia_env(quiet):
        remove_julia_env()
        if quiet:
            with suppress_output():
                juliapkg.resolve(force=True)
        else:
            juliapkg.resolve(force=True)

    try:
        _ensure_julia_deps_ready(quiet)
    except Exception:
        if retry:
            _reset_julia_env(quiet)
            _ensure_julia_deps_ready(quiet)
            return
        raise


def remove_julia_env():
    """Removes the active Julia environment directory.

    Notes
    -----
    When Julia or its dependencies are corrupted, this is a possible fix.
    """
    path_julia_env = Path(juliapkg.project())

    if path_julia_env.exists():
        log.warning(f"Removing Julia environment directory: {path_julia_env}")
        shutil.rmtree(path_julia_env)
    else:
        log.warning("Julia environment directory not found.")


# Lazy loading for jl and ec
class _LazyJuliaRuntime:
    def __init__(self):
        self._initialized = False
        self._jl = None
        self._ec = None

    def _init(self):
        if not self._initialized:
            log.info("Initializing Julia runtime...")
            os.environ["PYTHON_JULIACALL_AUTOLOAD_IPYTHON_EXTENSION"] = "no"
            ensure_julia_deps_ready(quiet=True)
            self._jl = init_julia(quiet=True)
            self._ec = import_backend(self._jl)
            self._initialized = True

    def get_jl(self):
        self._init()
        return self._jl

    def get_ec(self):
        self._init()
        return self._ec


class _LazyProxy:
    def __init__(self, getter):
        self._getter = getter

    def __getattr__(self, name):
        return getattr(self._getter(), name)

    def __call__(self, *args, **kwargs):
        return self._getter()(*args, **kwargs)

    def __repr__(self):
        return repr(self._getter())


_lazy_rt = _LazyJuliaRuntime()
jl = _LazyProxy(_lazy_rt.get_jl)
ec = _LazyProxy(_lazy_rt.get_ec)



"""
Collection of functions for calculating metrics.

.. currentmodule:: autoeis.metrics

.. autosummary::
   :toctree: generated/

    mape_score
    mse_score
    rmse_score
    r2_score

"""

import numpy as np


def _assert_compatible_shapes(y_true: np.ndarray, y_pred: np.ndarray, axis: int):
    # y_true must be 1D, but y_pred can be ND (ND y_pred is complicated)
    assert y_true.squeeze().ndim == 1, "y_true must be 1D."
    # This is not necessary (numpy catches it), but it's a good sanity check
    msg = (
        f"y_pred {y_pred.shape} is incompatibe with y_true {y_true.shape} "
        f"given axis = {axis}"
    )
    assert y_true.shape[axis] == y_pred.shape[axis], msg


def _reshape_given_axis(y_true: np.ndarray, y_pred: np.ndarray, axis: int):
    y_true = y_true.squeeze()
    # For broadcasting to work correctly, ensure y_true is expanded in the correct axis
    other_axes = [i for i in range(y_pred.ndim) if i != axis]
    y_true = np.expand_dims(y_true, axis=other_axes)
    return y_true


def mape_score(y_true: np.ndarray, y_pred: np.ndarray, axis=0):
    """
    Calculates the generalized MAPE (Mean Absolute Percentage Error) score.

    Parameters
    ----------
    y_true : np.ndarray
        Ground truth (true) values.
    y_pred : np.ndarray
        Predicted values.
    axis : int, optional
        Axis along which to calculate the MAPE score. Default is 0.

    Returns
    -------
    float
        The MAPE score as a percentage.

    Notes
    -----
    This function handles complex numbers in the input arrays.
    """
    y_true = _reshape_given_axis(y_true, y_pred, axis)
    _assert_compatible_shapes(y_true, y_pred, axis)
    # NOTE: abs is needed to handle complex numbers
    return np.mean(np.abs((y_true - y_pred) / y_true), axis=axis) * 100


def mse_score(y_true: np.ndarray, y_pred: np.ndarray, axis=0):
    """
    Calculates the generalized MSE (Mean Squared Error) score.

    Parameters
    ----------
    y_true : np.ndarray
        Ground truth (true) values.
    y_pred : np.ndarray
        Predicted values.
    axis : int, optional
        Axis along which to calculate the MSE score. Default is 0.

    Returns
    -------
    float
        The MSE score.

    Notes
    -----
    This function handles complex numbers in the input arrays.
    """
    y_true = _reshape_given_axis(y_true, y_pred, axis)
    _assert_compatible_shapes(y_true, y_pred, axis)
    # NOTE: abs is needed to handle complex numbers
    return np.mean(np.abs(y_true - y_pred) ** 2, axis=axis)


def rmse_score(y_true: np.ndarray, y_pred: np.ndarray, axis=0):
    """
    Calculates the generalized RMSE (Root Mean Squared Error) score.

    Parameters
    ----------
    y_true : np.ndarray
        Ground truth (true) values.
    y_pred : np.ndarray
        Predicted values.
    axis : int, optional
        Axis along which to calculate the RMSE score. Default is 0.

    Returns
    -------
    float
        The RMSE score.

    Notes
    -----
    This function handles complex numbers in the input arrays.
    """
    y_true = _reshape_given_axis(y_true, y_pred, axis)
    _assert_compatible_shapes(y_true, y_pred, axis)
    return np.sqrt(mse_score(y_true, y_pred, axis=axis))


def r2_score(y_true: np.ndarray, y_pred: np.ndarray, axis=0):
    """
    Calculates the generalized R2 score.

    Parameters
    ----------
    y_true : np.ndarray
        Ground truth (true) values.
    y_pred : np.ndarray
        Predicted values.
    axis : int, optional
        Axis along which to calculate the R2 score. Default is 0.

    Returns
    -------
    float
        The R2 score.

    Notes
    -----
    This function handles complex numbers in the input arrays.
    """
    y_true = _reshape_given_axis(y_true, y_pred, axis)
    _assert_compatible_shapes(y_true, y_pred, axis)
    # NOTE: abs is needed to handle complex numbers
    ssr = np.sum(np.abs(y_true - y_pred) ** 2, axis=axis)
    sst = np.sum(np.abs(y_true - np.mean(y_true)) ** 2)
    return 1 - ssr / sst



"""
Collection of functions core to AutoEIS functionality.

.. currentmodule:: autoeis.core

.. autosummary::
   :toctree: generated/

    perform_full_analysis
    generate_equivalent_circuits
    filter_implausible_circuits
    perform_bayesian_inference

"""

import logging
import os
import time
import warnings
from collections.abc import Iterable, Mapping
from copy import deepcopy

import arviz as az
import jax
import jax.numpy as jnp  # noqa: F401
import numpy as np
import numpyro
import pandas as pd
import psutil
from deprecated import deprecated
from jax import config
from mpire import WorkerPool
from numpyro.distributions import Distribution
from numpyro.infer import MCMC, NUTS
from scipy.optimize import curve_fit
from tqdm.auto import tqdm

from autoeis import io, julia_helpers, metrics, models, parser, utils
from autoeis.julia_helpers import ec, jl

from .utils import InferenceResult

# FP32 gradients not sufficient for ECM fitting
numpyro.enable_x64()
# GPU is actually slower for typical EIS dataset sizes
numpyro.set_platform("cpu")
# Get rid of the JAX warning about GPU might be available
config.update("jax_platforms", "cpu")

warnings.filterwarnings("ignore", category=Warning, module="arviz.*")
log = logging.getLogger(__name__)

__all__ = [
    "perform_full_analysis",
    "generate_equivalent_circuits",
    "filter_implausible_circuits",
    "perform_bayesian_inference",
    "compute_fitness_metrics",
    "jl",
    "ec",
]


# TODO: Breaks when data is noisy -> use curve_fit to extrapolate R0
def compute_ohmic_resistance(freq: np.ndarray[float], Z: np.ndarray[complex]) -> float:
    """Extracts the ohmic resistance from impedance data.

    Parameters
    ----------
    Z : np.ndarray[complex]
        Impedance measurements.
    freq : np.ndarray[float]
        Frequencies corresponding to impedance measurements.

    Returns
    -------
    ohmic_resistance: float
        The ohmic resistance of impedance data.

    Raises
    ------
    ValueError
        If the ohmic resistance cannot be reliably extracted.
    """
    # Sort impedance data by descending frequency
    mask = np.argsort(freq)[::-1]
    Z = Z[mask]
    freq = freq[mask]

    # Curve fit to a saturating function: (ax + b) / (x + c)
    # NOTE: x -> inf, f(x) -> a === 1/R0 => set a > 0 as lower bound
    # NOTE: To put more weight on high frequency data, use 1/Z.real as y
    x = freq[:]
    y = 1 / Z.real[:]

    def func(x, a, b, c):
        return (a * x + b) / (x + c)

    fallback = False
    bounds = ([0, -np.inf, -np.inf], np.inf)

    try:
        (a, b, c), pcov = curve_fit(func, x, y, bounds=bounds)
        R = 1 / a
    except RuntimeError:
        fallback = True
    if R < 0 or R < 0.2 * Z.real[0]:
        fallback = True
    if fallback:
        R = Z.real[0]
        log.warning("Failed to fit ohmic resistance, returning Re(Z) @ max(freq).")
    return R


def generate_equivalent_circuits(
    freq: np.ndarray[float],
    Z: np.ndarray[complex],
    iters: int = 100,
    complexity: int = 12,
    tol: float = 1e-2,
    parallel: bool = True,
    generations: int = 30,
    population_size: int = 100,
    terminals: str = "RLP",
    seed: int = None,
) -> pd.DataFrame:
    """Generates candidate circuits that fit the impedance data using
    evolutionary algorithms.

    Parameters
    ----------
    freq : np.ndarray[float]
        Frequencies corresponding to impedance measurements.
    Z : np.ndarray[complex]
        Impedance measurements as a complex array.
    iters : int, optional
        Number of ECM generation iterations (default is 100).
    complexity : int, optional
        Complexity of the ECM search space (default is 12).
    tol : float, optional
        Convergence threshold for the ECM search (default is 1e-32).
    parallel : bool, optional
        If True, the ECM search will be performed in parallel (default is True).
    generations : int, optional
        Number of generations for the ECM search (default is 30).
    population_size : int, optional
        Number of ECMs to generate per generation (default is 100).
    terminals : str, optional
        Circuit components to consider (default is "RLP"). R: resistor,
        L: inductor, C: capacitor, P: constant-phase element.
    seed : int, optional
        Random seed for reproducibility (default is None).

    Returns
    -------
    pd.DataFrame or None
        Dataframe containing circuits.
    """
    log.info("Generating equivalent circuits via evolutionary algorithms.")

    # Set the seed for reproducibility (if not set, use current time in nanoseconds)
    seed = seed or time.time_ns() % 2**32

    ec_kwargs = {
        "head": complexity,
        "terminals": terminals,
        "convergence_threshold": tol,
        "generations": generations,
        "population_size": population_size,
    }

    ecm_generator = _generate_ecm_parallel_julia if parallel else _generate_ecm_serial
    circuits = ecm_generator(freq, Z, iters, ec_kwargs, seed)

    # Convert output to DataFrame with columns ("circuitstring", "Parameters")
    circuits = io.parse_ec_output(circuits)

    if not len(circuits):
        log.warning("No plausible circuits found. Increase 'tol', 'iters', or both!")

    return circuits


def _generate_ecm_serial(
    freq: np.ndarray[float],
    Z: np.ndarray[complex],
    iters: int,
    ec_kwargs: dict,
    seed: int,
) -> list[str]:
    """Generates candidate circuits that fit the impedance data, in serial."""
    # Set random seed for reproducibility
    jl.seval(f"import Random; Random.seed!({seed})")

    circuits = []
    for _ in tqdm(range(iters), desc="Generating Candidate ECMs", leave=False):
        utils.flush_streams()
        try:
            circuit = ec.circuit_evolution(Z, freq, **ec_kwargs, quiet=True)
        except Exception as e:
            log.error(f"Error generating circuit: {e}")
            continue
        circuits.append(circuit)
    else:
        utils.flush_streams()

    circuits = [str(c) for c in circuits if c is not None]
    return circuits


def _generate_ecm_parallel_julia(
    freq: np.ndarray[float],
    Z: np.ndarray[complex],
    iters: int,
    ec_kwargs: dict,
    seed: int,
):
    """Generates candidate circuits that fit the impedance data, in parallel
    via Julia multiprocessing.
    """
    # Set random seed for reproducibility (Python and Julia)
    # FIXME: This doesn't work when multiprocessing, use @everywhere instead
    jl.seval(f"import Random; Random.seed!({seed})")

    # HACK: To get a progress bar, chunk the iterations -> call Julia repeatedly
    nprocs = psutil.cpu_count(logical=False)
    # Double the number of workers to buffer for those with slow convergence
    # (don't do this for small iters, otherwise progress bar becomes pointless)
    nprocs = 2 * nprocs if (iters // nprocs) > 10 else nprocs
    # NOTE: e.g., iters = 11, nprocs = 4 -> iters_chunked = [4, 4, 3]
    iters_chunked = [nprocs] * (iters // nprocs)
    if iters % nprocs:
        iters_chunked.append(iters % nprocs)

    # Perform parallelized GEP in chunks
    circuits = []

    with tqdm(total=iters, desc="Generating Candidate ECMs", miniters=1, leave=False) as pbar:
        utils.flush_streams()
        for iters_ in iters_chunked:
            try:
                circuits_ = ec.circuit_evolution_batch(
                    Z, freq, **ec_kwargs, iters=iters_, quiet=True
                )
            except Exception as e:
                log.error(f"Error generating circuits: {e}")
                circuits_ = []
            circuits += circuits_
            pbar.update(iters_)
            utils.flush_streams()

    circuits = [str(c) for c in circuits if c is not None]
    return circuits


@deprecated(reason="This function is deprecated, use _generate_ecm_parallel_julia instead")
def _generate_ecm_parallel_mpire(
    freq: np.ndarray[float],
    Z: np.ndarray[complex],
    iters: int,
    ec_kwargs: dict,
    seed: int,
):
    """Generates candidate circuits that fit the impedance data, in parallel
    via Python multiprocessing.
    """

    def circuit_evolution(seed: int):
        """Closure to generate a single circuit to be used with multiprocessing."""
        jl = julia_helpers.init_julia()
        ec = julia_helpers.import_backend(jl)
        # Set random seed for reproducibility
        jl.seval(f"import Random; Random.seed!({seed})")
        try:
            circuit = ec.circuit_evolution(Z, freq, **ec_kwargs)
        except Exception as e:
            log.error(f"Error generating circuit: {e}")
            return None
        return circuit

    nproc = os.cpu_count()
    mpire_kwargs = {
        "iterable_len": iters,
        "progress_bar": True,
        "progress_bar_style": "notebook",
        "progress_bar_options": {"desc": "Circuit Evolution"},
    }

    # Julia cannot be initialized in the main process -> guard against this
    runtime_error = False

    # Set a different seed for each process
    seed = [seed + i for i in range(iters)]

    with WorkerPool(n_jobs=nproc) as pool:
        try:
            circuits = pool.map(circuit_evolution, seed, **mpire_kwargs)
        except RuntimeError:
            runtime_error = True

    if runtime_error:
        raise RuntimeError("Julia must not be manually initialized, restart the kernel.")

    circuits = [str(c) for c in circuits if c is not None]
    return circuits


def split_components(circuits: pd.DataFrame) -> pd.DataFrame:
    """Augments the circuits dataframe with additional columns for each component."""
    # Initialize lists to populate the component columns later
    components = {"R": [], "C": [], "L": [], "P": []}
    labels = {"R": "Resistors", "C": "Capacitors", "L": "Inductors", "P": "CPEs"}

    for row in circuits.itertuples():
        circuit = row.circuitstring
        # Find components of each kind
        pgroups = parser.group_parameters_by_component(circuit)
        for ctype in components.keys():
            components[ctype].append(pgroups.get(ctype, []))

    # Add component columns to the dataframe
    for ctype, params in components.items():
        column_label = labels[ctype]
        circuits[column_label] = params

    return circuits


def capacitance_filter(circuits: pd.DataFrame) -> pd.DataFrame:
    """Excludes ideal capacitors from the circuits dataframe."""
    circuits = circuits.copy(deep=True)

    for row in circuits.itertuples():
        variables = row.Parameters.keys()
        contains_capacitor = any("C" in var for var in variables)
        if contains_capacitor:
            circuits.drop(row.Index, inplace=True)
    circuits.reset_index(drop=True, inplace=True)

    return circuits


def ohmic_resistance_filter(circuits: pd.DataFrame) -> pd.DataFrame:
    """Excludes circuits without an ohmic resistance from the circuits dataframe."""
    circuits = circuits.copy(deep=True)

    for row in circuits.itertuples():
        circuit = row.circuitstring
        resistors = parser.find_ohmic_resistors(circuit)
        if not resistors:
            circuits.drop(row.Index, inplace=True)

    circuits.reset_index(drop=True, inplace=True)
    return circuits


def series_filter(circuits: pd.DataFrame) -> pd.DataFrame:
    """Excludes circuits with series-only components from the circuits dataframe."""
    circuits = circuits.copy(deep=True)

    for row in circuits.itertuples():
        circuit = row.circuitstring
        contains_parallel_route = "[" in circuit
        if not contains_parallel_route:
            circuits.drop(row.Index, inplace=True)

    circuits.reset_index(drop=True, inplace=True)
    return circuits


def merge_identical_circuits(circuits: "pd.DataFrame") -> "pd.DataFrame":
    """Merges identical circuits from the circuits dataframe."""
    circuits = circuits.copy(deep=True)

    for i, row_i in circuits.iterrows():
        circuit_i = row_i.circuitstring
        for j, row_j in circuits.loc[i + 1 :].iterrows():
            if utils.are_circuits_equivalent(circuit_i, row_j.circuitstring):
                circuits.drop(j, inplace=True)

    circuits.reset_index(drop=True, inplace=True)
    return circuits


def compute_fitness_metrics(
    circuits: pd.DataFrame, freq: np.ndarray[float], Z: np.ndarray[complex]
) -> pd.DataFrame:
    """Computes various fitness metrics and returns an augmented dataframe.

    Parameters
    ----------
    circuits : pd.DataFrame
        Circuits dataframe with inference results
    freq : np.ndarray[float]
        Frequencies corresponding to the impedance data
    Z : np.ndarray[complex]
        Complex impedance data

    Returns
    -------
    circuits : pd.DataFrame
        Circuits dataframe with fitness metrics
    """
    circuits = circuits.copy(deep=True)

    results = circuits["InferenceResult"]
    mcmcs = [result.mcmc for result in results]
    circuits["converged"] = [result.converged for result in results]
    circuits["divergences"] = [result.num_divergences for result in results]

    # Compute the posterior predictive and add to the dataframe
    Z_pred = [utils.eval_posterior_predictive(r.samples, r.circuit, freq) for r in results]
    circuits["Z_pred"] = Z_pred

    # FIXME: This function doesn't work with Bode as objective function
    # TODO: Extract log-likelihood of mag/phase, and turn it into that for real/imag

    # Compute WAIC and add to the dataframe
    if "sigma.real" in mcmcs[0].get_samples().keys():
        waic_re = [az.waic(x, var_name="obs.real", scale="deviance") for x in mcmcs]
        waic_im = [az.waic(x, var_name="obs.imag", scale="deviance") for x in mcmcs]
        circuits["WAIC (real)"] = [x["elpd_waic"] + 2 * x["p_waic"] for x in waic_re]
        circuits["WAIC (imag)"] = [x["elpd_waic"] + 2 * x["p_waic"] for x in waic_im]
    else:
        waic_mag = [az.waic(x, var_name="obs.mag", scale="deviance") for x in mcmcs]
        waic_phase = [az.waic(x, var_name="obs.phase", scale="deviance") for x in mcmcs]
        circuits["WAIC (mag)"] = [x["elpd_waic"] + 2 * x["p_waic"] for x in waic_mag]
        circuits["WAIC (phase)"] = [x["elpd_waic"] + 2 * x["p_waic"] for x in waic_phase]

    # Compute R^2 and add to the dataframe
    fn = lambda row: metrics.r2_score(Z.real, row.Z_pred.real, axis=1)
    circuits["R^2 (real)"] = circuits.apply(fn, axis=1)
    fn = lambda row: metrics.r2_score(Z.imag, row.Z_pred.imag, axis=1)
    circuits["R^2 (imag)"] = circuits.apply(fn, axis=1)
    # Since R^2 is a vector (num_samples), also compute its mean for convenience
    circuits["R^2 (ravg)"] = circuits["R^2 (real)"].apply(np.mean)
    circuits["R^2 (iavg)"] = circuits["R^2 (imag)"].apply(np.mean)

    # Compute MAPE and add to the dataframe
    fn = lambda row: metrics.mape_score(Z.real, row.Z_pred.real, axis=1)
    circuits["MAPE (real)"] = circuits.apply(fn, axis=1)
    fn = lambda row: metrics.mape_score(Z.imag, row.Z_pred.imag, axis=1)
    circuits["MAPE (imag)"] = circuits.apply(fn, axis=1)
    # Since MAPE is a vector (num_samples), also compute its mean for convenience
    circuits["MAPE (ravg)"] = circuits["MAPE (real)"].apply(np.mean)
    circuits["MAPE (iavg)"] = circuits["MAPE (imag)"].apply(np.mean)

    # Add number of parameters to the dataframe
    circuits["n_params"] = circuits.apply(lambda row: len(row.Parameters), axis=1)

    return circuits


def _validate_circuit(circuits):
    """Validates the circuits input."""
    # Dataframe
    if isinstance(circuits, pd.DataFrame):
        utils.validate_circuits_dataframe(circuits)
    # List
    elif isinstance(circuits, list):
        for c in circuits:
            assert isinstance(c, str), f"Invalid circuit: {c}"
    # String
    elif isinstance(circuits, str):
        pass
    else:
        raise ValueError("`circuits` must be a DataFrame, list, or string.")
    return circuits


def _validate_impedance_data(freq, Z):
    """Validates the impedance data input."""
    # Shared validations (single/multiple datasets)
    assert type(freq) is type(Z), "freq and Z must be of the same type."
    assert utils.is_iterable(freq), "freq must be an ndarray or list/tuple[ndarray]."
    assert utils.is_iterable(Z), "Z must be an ndarray or list/tuple[ndarray]."
    assert len(freq) == len(Z), "freq and Z must have the same length."
    # Standardize freq/Z to be an np.ndarray (freq/Z[i] for multiple datasets)
    if utils.is_nested_iterable(freq):
        for i, (f, z) in enumerate(zip(freq, Z)):
            assert isinstance(f, np.ndarray), f"freq must be a numpy array (dataset #{i})."
            assert isinstance(z, np.ndarray), f"Z must be a numpy array (dataset #{i})."
            assert f.size == z.size, f"freq and Z must have the same length (dataset #{i})."
    else:
        assert isinstance(freq, np.ndarray), "freq must be a numpy array."
        assert isinstance(Z, np.ndarray), "Z must be a numpy array."
    return freq, Z


def _validate_p0(p0, broadcast_to: int = 1):
    """Validates the initial guess for ECM parameters, and broadcasts to a list."""
    # Validate input data type
    if not (isinstance(p0, (dict, list)) or utils.is_ndfarray_like(p0) or p0 is None):
        raise ValueError("'p0' must be an ndarray/dict or list[ndarray/dict].")

    # If single p0 is provided, broadcast
    if p0 is None or utils.is_ndfarray_like(p0) or isinstance(p0, dict):
        p0 = [deepcopy(p0) for _ in range(broadcast_to)]

    # If a list of p0 is provided, validate
    assert isinstance(p0, list), "Something went wrong. Please report!"
    for p0_ in p0:
        msg = f"Invalid p0: {p0_}"
        assert utils.is_ndfarray_like(p0_) or isinstance(p0_, dict), msg

    return p0


def _validate_priors(priors, broadcast_to: int = 1):
    """Validates the priors for the ECM parameters, and broadcasts to a list."""
    # Validate input data type
    if not isinstance(priors, (Mapping, list, type(None))):
        raise ValueError("'priors' must be a dict or list[dict].")

    # If single priors is provided, broadcast
    if priors is None or isinstance(priors, Mapping):
        priors = [deepcopy(priors) for _ in range(broadcast_to)]

    # Ensure all priors are valid
    assert isinstance(priors, list), "Something went wrong. Please report!"
    for priors_ in priors:
        msg = "'priors' must be a dict of NumPyro Distribution objects."
        assert isinstance(priors_, Mapping), msg
        assert all(isinstance(v, Distribution) for v in priors_.values()), msg

    return priors


def _validate_seed(seed, num_splits=1) -> list[jax.Array] | jax.Array:
    """Validates the random seed to be used in ``MCMC.run()``."""
    # If seed is not set, generate a new seed using time
    if seed is None:
        seed = time.time_ns() % 2**32
    # If a jax.Array is already provided, use it otherwise generate one
    if isinstance(seed, jax.Array):
        assert seed.size == 2, "'seed' must be a 2-element jax.Array."
        subkey = seed
    elif isinstance(seed, int):
        key = jax.random.PRNGKey(seed)
        key, *subkey = jax.random.split(key, num_splits + 1)
        subkey = subkey[0] if num_splits == 1 else subkey
    else:
        raise ValueError("'seed' must be an int or a jax.Array.")
    return subkey


def _refine_p0(p0, circuit, datasets, progress_bar):
    """"""
    if p0 is None:
        p0 = [None] * len(datasets)
    if len(p0) != len(circuit) != len(datasets):
        raise ValueError("Length of 'p0', 'circuit', and 'datasets' must be the same.")

    max_iters = 10
    p0 = utils.distribute_task(
        utils.fit_circuit_parameters,
        circuit,
        [dataset.freq for dataset in datasets],
        [dataset.Z for dataset in datasets],
        p0,
        max_iters,
        static=(4),  # static args = (circuit, freq, Z, max_iters)
        progress_bar=progress_bar,
        desc="Refining Initial Guess",
    )

    # Handle failed refinements
    for i, elem in enumerate(p0):
        if isinstance(elem, Exception):
            raise RuntimeError(f"Failed to refine p0 for circuit {circuit[i]}: {elem}")

    return p0


def perform_bayesian_inference(
    circuit: str | Iterable[str] | pd.DataFrame,
    freq: np.ndarray[float] | Iterable[float],
    Z: np.ndarray[complex] | Iterable[complex],
    p0: Iterable[float]
    | Mapping[str, float]
    | Iterable[Iterable[float]]
    | Iterable[Mapping[str, float]] = None,
    priors: Mapping[str, Distribution] = None,
    num_warmup: int = 2500,
    num_samples: int = 1000,
    num_chains: int = 1,
    seed: int | jax.Array = None,
    method: str = "bode",
    progress_bar: bool = True,
    refine_p0: bool = True,
    parallel: bool = True,
) -> list[InferenceResult]:
    """Performs Bayesian inference on the circuits based on impedance data.

    Parameters
    ----------
    circuit : str | Iterable[str] | pd.DataFrame
        Dataframe containing circuits, list of circuit strings, or a single
        circuit string.
    freq: np.ndarray[float] | Iterable[float]
        Frequency array corresponding to the impedance data. If a list of
        frequencies is provided, separate inferences will be performed for
        each frequency array.
    Z : np.ndarray[complex] | Iterable[np.ndarray[complex]]
        Impedance data as a complex array. If a list of impedance arrays is
        provided, separate inferences will be performed for each impedance
        array.
    p0 : Iterable[float] \
            | Mapping[str, float] \
            | Iterable[Iterable[float]] \
            | Iterable[Mapping[str, float]], optional
        Initial guess for the circuit parameters (default is None).
    priors : Mapping[str, Distribution], optional
        Priors for the circuit parameters (default is None).
    num_warmup : int, optional
        Number of warmup samples for the MCMC (default is 2500).
    num_samples : int, optional
        Number of samples for the MCMC (default is 1000).
    num_chains : int, optional
        Number of MCMC chains (default is 1).
    seed : int, optional
        Random seed for reproducibility (default is None).
    method : str, optional
        Objective function to minimize (default is "bode"). Choose from "bode",
        "nyquist", "magnitude", or "chi-squared".
    progress_bar : bool, optional
        If True, a progress bar will be displayed (default is True).
    refine_p0 : bool, optional
        If True, the initial guess for the circuit parameters will be refined
        using the circuit fitter (default is True).
    parallel : bool, optional
        If True, the MCMC chains will be run in parallel (default is True).

    Returns
    -------
    list[InferenceResult]
        List of InferenceResult object(s). Regardless of whether a single or
        multiple circuits/datasets are provided, the output will always be a
        list of InferenceResult objects for consistency.

    Notes
    -----
    Only the following combinations of inputs are supported:
    - Single circuit, single dataset
    - Single circuit, multiple datasets
    - Multiple circuits, single dataset

    """
    # Validate inputs data types
    circuit = _validate_circuit(circuit)
    freq, Z = _validate_impedance_data(freq, Z)
    # Convert dataframe to list if necessary
    if isinstance(circuit, pd.DataFrame):
        p0 = circuit["Parameters"].tolist()
        circuit = circuit["circuitstring"].tolist()

    # Standardize circuits and impedance data to always be a list
    circuit = [circuit] if isinstance(circuit, str) else circuit
    freq, Z = ([freq], [Z]) if isinstance(freq, np.ndarray) else (freq, Z)
    datasets = [utils.ImpedanceData(freq_, Z_) for freq_, Z_ in zip(freq, Z)]

    # Determine processing workflow: SCSD, SCMD, MCSD
    num_inferences = max(len(circuit), len(datasets))
    # Raise error if multiple circuits and datasets are provided together
    if len(circuit) > 1 and len(datasets) > 1:
        raise ValueError("Can't handle multiple circuits and multiple datasets together.")
    # Ensure circuits and datasets are of the same length (if not, broadcast)
    circuit = circuit * num_inferences if len(circuit) == 1 else circuit
    datasets = datasets * num_inferences if len(datasets) == 1 else datasets

    # Ensure p0 and priors are valid and ready for parallel inference
    p0 = _validate_p0(p0, broadcast_to=num_inferences)
    if refine_p0 or (p0[0] is None):
        p0 = _refine_p0(p0, circuit, datasets, progress_bar)
    if priors is None:
        priors = [utils.initialize_priors(p0_) for p0_ in p0]
    priors = _validate_priors(priors, broadcast_to=num_inferences)

    # Generate N random seeds (one for each inference)
    seed = _validate_seed(seed, num_splits=num_inferences)
    # Ensure seed is a list since MCMD requires all inputs to be lists
    seed = [seed] if not isinstance(seed, list) else seed

    # Perform Bayesian inference
    if parallel and num_inferences > 1:
        results = _perform_bayesian_inference_MCMD(
            circuit=circuit,
            dataset=datasets,
            priors=priors,
            num_warmup=num_warmup,
            num_samples=num_samples,
            num_chains=num_chains,
            method=method,
            seed=seed,
            progress_bar=progress_bar,
        )
    else:
        results = []
        for i in tqdm(range(num_inferences), desc="Running Bayesian Inference", leave=False):
            result = _perform_bayesian_inference_SCSD(
                circuit=circuit[i],
                freq=datasets[i].freq,
                Z=datasets[i].Z,
                priors=priors[i],
                num_warmup=num_warmup,
                num_samples=num_samples,
                num_chains=num_chains,
                seed=seed[i],
                method=method,
                progress_bar=False,
            )
            results.append(result)

    return results


def _perform_bayesian_inference_SCSD(
    circuit: str,
    freq: np.ndarray[float],
    Z: np.ndarray[complex],
    priors: Mapping[str, Distribution],
    num_warmup: int,
    num_samples: int,
    num_chains: int,
    method: str,
    seed: int | jax.Array,
    progress_bar: bool = False,
) -> InferenceResult:
    log.info("Performing Bayesian inference on the circuit {circuit}.")

    subkey = _validate_seed(seed)
    circuit_fn = utils.generate_circuit_fn(circuit, jit=True)
    method = method.replace("-", "_")

    nuts_kernel = NUTS(
        model=getattr(models, f"circuit_regression_{method}"),
        init_strategy=numpyro.infer.init_to_median,
    )
    kwargs_mcmc = {
        "num_samples": num_samples,
        "num_warmup": num_warmup,
        "num_chains": num_chains,
        "progress_bar": progress_bar,
    }
    mcmc = MCMC(nuts_kernel, **kwargs_mcmc)
    kwargs_inference = {
        "freq": freq,
        "Z": Z,
        "priors": priors,
        "circuit_fn": circuit_fn,
    }

    try:
        # HACK: Handle when circuit fitter fails to find p0 (to be caught by RuntimeError)
        if priors is None:
            raise RuntimeError(
                f"Inference couldn't be performed for circuit: {circuit}, "
                "because 'priors' are not provided, possibly because "
                "circuit fitter failed to find an initial guess 'p0'."
            )
        mcmc.run(subkey, **kwargs_inference)
        converged = True
    except RuntimeError as e:
        log.error(f"Inference failed for circuit: {circuit}. Error: {e}")
        converged = False

    return InferenceResult(circuit, mcmc, converged=converged, freq=freq, Z=Z)


def _perform_bayesian_inference_MCMD(
    circuit: Iterable[str],
    dataset: Iterable[utils.ImpedanceData],
    priors: Iterable[Mapping[str, Distribution]],
    num_warmup: int,
    num_samples: int,
    num_chains: int,
    method: str,
    seed: jax.Array,
    progress_bar: bool,
) -> list[InferenceResult]:
    """Performs inference in batch mode; multiple-circuits multiple-data (MCMD)."""
    # Sanity check on input lengths
    msg = "'circuit', 'dataset', and 'priors' must have the same length."
    assert len(circuit) == len(dataset) == len(priors), msg

    # Fetch freq, Z pairs from the dataset
    freq, Z = zip(*[(dataset_.freq, dataset_.Z) for dataset_ in dataset])

    # If a single inference is requested, perform SCSD directly to avoid overhead
    if len(circuit) == 1:
        results = _perform_bayesian_inference_SCSD(
            circuit[0],
            freq[0],
            Z[0],
            priors=priors[0],
            num_warmup=num_warmup,
            num_samples=num_samples,
            num_chains=num_chains,
            method=method,
            seed=seed[0],
            progress_bar=False,
        )
        results = [results]
    else:
        results = utils.distribute_task(
            _perform_bayesian_inference_SCSD,
            # --- _perform_bayesian_inference_SCSD args ---
            circuit,  # 0
            freq,  # 1
            Z,  # 2
            priors,  # 3
            num_warmup,  # 4
            num_samples,  # 5
            num_chains,  # 6
            method,  # 7
            seed,  # 8
            # --- distribute_task kwargs ---
            static=(4, 5, 6, 7),  # indices of static args
            progress_bar=progress_bar,
            desc="Performing Bayesian Inference",
        )
    return results


def filter_implausible_circuits(circuits: pd.DataFrame) -> pd.DataFrame:
    """Applies heuristic rules to exclude implausible circuits.

    Parameters
    ----------
    circuits : pd.DataFrame
        Dataframe containing circuits.

    Returns
    -------
    circuits : pd.DataFrame
        Dataframe containing the filtered circuits.
    """
    log.info("Filtering the circuits using heuristic rules.")
    circuits = circuits.copy(deep=True)

    if len(circuits) == 0:
        log.warning("Circuits' dataframe is empty!")
        return circuits

    circuits = split_components(circuits)
    circuits = capacitance_filter(circuits)
    circuits = series_filter(circuits)
    circuits = ohmic_resistance_filter(circuits)
    circuits = merge_identical_circuits(circuits)

    # Drop the columns added by split_components
    circuits = circuits.drop(columns=["Resistors", "Capacitors", "Inductors", "CPEs"])

    if len(circuits) == 0:
        log.warning("No plausible circuits left after post-filtering")  # fmt: skip

    return circuits


def perform_full_analysis(
    freq: np.ndarray[float],
    Z: np.ndarray[complex],
    iters: int = 100,
    parallel: bool = True,
    tol_linKK: float = 5e-2,
    tol: float = 1e-2,
    num_warmup: int = 2500,
    num_samples: int = 1000,
) -> pd.DataFrame:
    """Performs automated EIS analysis by generating plausible ECMs that
    fit the impedance data, followed by Bayesian inference on components.

    Parameters
    ----------
    freq : np.ndarray[float]
        Frequencies corresponding to the impedance data.
    Z : np.ndarray[complex]
        Impedance data as a complex array.
    iters : int, optional
        Number of iterations for ECM generation. Default is 100.
    parallel : bool, optional
        If True, the ECM generation will be done in parallel. Default is True.
    tol_linKK : float, optional
        Tolerance for acceptable measurements based on linKK residuals.
    tol : float, optional
        Convergence threshold for the ECM generation. Default is 1e-2.
    num_warmup : int, optional
        Number of warmup samples for the MCMC. Default is 2500.
    num_samples : int, optional
        Number of samples for the MCMC. Default is 1000.

    Returns
    -------
    circuits: pd.DataFrame
        Dataframe containing circuits, parameters, and inference results.
    """
    raise NotImplementedError(
        "This function has some issues and is not yet ready to use. "
        "Regardless, it is recommended to use the step-by-step approach for "
        "more control. Please refer to the documentation for more details."
    )

    # Filter out bad impedance data
    freq, Z = utils.preprocess_impedance_data(freq, Z, tol_linKK=tol_linKK)

    # Generate a pool of potential ECMs via an evolutionary algorithm
    kwargs = {"iters": iters, "complexity": 12, "tol": tol, "parallel": parallel}
    circuits_unfiltered = generate_equivalent_circuits(freq, Z, **kwargs)

    # Apply heuristic rules to filter unphysical circuits
    circuits = filter_implausible_circuits(circuits_unfiltered)

    # Perform Bayesian inference on the filtered ECMs
    kwargs_mcmc = {"num_warmup": num_warmup, "num_samples": num_samples}
    results = perform_bayesian_inference(circuits, freq, Z, **kwargs_mcmc)

    # Add the results to the circuits dataframe as a new column
    circuits["MCMC"] = [result.mcmc for result in results]
    circuits["success"] = [result.converged for result in results]
    circuits["divergences"] = [result.num_divergences for result in results]

    # Compute fitness metrics and add to the dataframe
    circuits = compute_fitness_metrics(circuits, freq, Z)

    return circuits



from pathlib import Path

import click

from .julia_helpers import install_backend, install_julia


@click.group("autoeis")
@click.pass_context
def autoeis_installer(context):
    ctx = context


@click.option(
    "--ec-path",
    default=None,
    type=str,
    help="Installs a local copy of EquivalentCircuits instead of the remote version.",
)
@click.option(
    "--verbose",
    is_flag=True,
    help="Prints the installation process to the console.",
)
@autoeis_installer.command("install", help="Install Julia dependencies for AutoEIS.")
def install_cli(ec_path, verbose):
    if ec_path is not None:
        # Expand ~ in path if present
        ec_path = Path(ec_path).expanduser()
    install_julia(quiet=not verbose)
    install_backend(ec_path=ec_path, quiet=not verbose)



import logging

from rich.console import Console
from rich.logging import RichHandler

from .utils import Settings as _Settings


def _setup_logger():
    """Sets up logging using ``rich``."""
    log = logging.getLogger("autoeis")

    if log.hasHandlers():
        log.critical("Logging already set up.")
        return

    log.setLevel(logging.WARNING)
    console = Console(force_jupyter=False)
    handler = RichHandler(
        rich_tracebacks=True, console=console, show_path=not config.notebook
    )
    handler.setFormatter(logging.Formatter("%(message)s", datefmt="[%X]"))
    log.addHandler(handler)


config = _Settings()
_setup_logger()

from . import core, io, metrics, parser, utils, visualization  # noqa: F401, E402
from .core import *  # noqa: E402
from .version import __equivalent_circuits_jl_version__, __version__  # noqa: F401, E402
from .visualization import rich_print  # noqa: F401, E402



"""
Collection of functions to be used as models for Bayesian inference.

.. currentmodule:: autoeis.models

.. autosummary::
   :toctree: generated/

    circuit_regression_magnitude
    circuit_regression_nyquist
    circuit_regression_bode

Notes
-----

1. For positive variable `y`, there are multiple options to model the observation
to ensure that the predicted value is also positive:

- Sample ``y`` with a shifted ``HalfNormal`` (see pyro-ppl/numpyro/issues/932)
- Sample ``y`` with ``TruncatedNormal`` with 0 as the lower bound
- Sample ``y - y_gt`` with ``Normal(0, sigma)``

2. Using ``Normal`` distribution for positive variables may lead to negative
samples

"""

import copy
from collections import namedtuple
from typing import Callable, Mapping

import jax.numpy as jnp
import numpy as np
import numpyro
import numpyro.distributions as dist
from numpyro.distributions import Distribution

Impedance = namedtuple("Impedance", ["real", "imag"])


def circuit_regression_magnitude(
    freq: np.ndarray[float],
    priors: Mapping[str, Distribution],
    circuit_fn: Callable[[np.ndarray | float, np.ndarray], np.ndarray[complex]],
    Z: np.ndarray[complex] = None,
):
    """Inference model for ECM parameters based on |Z|.

    This model infers the circuit parameters by matching the magnitude of the
    impedance measurements. To do this, the MCMC sampling is performed on the
    magnitude of the impedance.

    Parameters
    ----------
    freq : np.ndarray[float]
        Frequencies to evaluate the posterior predictive distribution at.
    priors : Mapping[str, Distribution]
        Priors for the circuit parameters as a dictionary of parameter labels
        and distributions.
    circuit_fn : Callable[[np.ndarray | float, np.ndarray], np.ndarray[complex]]
        Function to compute the circuit impedance, parameterized by frequency
        and circuit parameters, fn(freq, p).
    Z : np.ndarray[complex], optional
        Observed impedance data. If ``None``, the model will be used for
        posterior prediction, otherwise it will be used for Bayesian
        inference of circuit parameters.
    """
    # Make a deep copy of the priors to avoid side effects (pyro-ppl/numpyro/issues/1651)
    priors = copy.deepcopy(priors)
    # Sample each element of X separately
    p = jnp.array([numpyro.sample(k, v) for k, v in priors.items()])
    # Predict Z using the model
    Z_pred = circuit_fn(freq, p)

    # Short-circuit if posterior prediction is requested
    if Z is None:
        _posterior_predictive(Z_pred)
        return

    # Observation model based on the magnitude of the impedance
    error_model = "abs(diff)"
    assert error_model in ["diff(abs)", "abs(diff)"]
    sigma = numpyro.sample("sigma", dist.Exponential(rate=1.0))
    if error_model == "diff(abs)":  # |Z| - |Z_gt|
        error = jnp.abs(Z) - jnp.abs(Z_pred)
        numpyro.sample("obs", dist.Normal(0.0, sigma), obs=error)
    if error_model == "abs(diff)":  # |Z - Z_gt|
        error = jnp.abs(Z - Z_pred)
        numpyro.sample("obs", dist.HalfNormal(sigma), obs=error)


def circuit_regression_nyquist(
    freq: np.ndarray[float],
    priors: Mapping[str, Distribution],
    circuit_fn: Callable[[np.ndarray | float, np.ndarray], np.ndarray[complex]],
    Z: np.ndarray[complex] = None,
):
    """Inference model for ECM parameters based on Nyquist plot.

    This model infers the circuit parameters by matching the impedance
    measurements as plotted on the Nyquist plot. To do this, the MCMC sampling
    is performed on the real and imaginary parts of the impedance.

    Parameters
    ----------
    freq : np.ndarray[float]
        Frequencies to evaluate the posterior predictive distribution at.
    priors : Mapping[str, Distribution]
        Priors for the circuit parameters as a dictionary of parameter labels
        and distributions.
    circuit_fn : Callable[[np.ndarray | float, np.ndarray], np.ndarray[complex]]
        Function to compute the circuit impedance, parameterized by frequency
        and circuit parameters, fn(freq, p).
    Z : np.ndarray[complex], optional
        Observed impedance data. If ``None``, the model will be used for
        posterior prediction, otherwise it will be used for Bayesian
        inference of circuit parameters.
    """
    # Make a deep copy of the priors to avoid side effects (pyro-ppl/numpyro/issues/1651)
    priors = copy.deepcopy(priors)
    # Sample each element of X separately
    p = jnp.array([numpyro.sample(k, v) for k, v in priors.items()])
    # Predict Z using the model
    Z_pred = circuit_fn(freq, p)

    # Short-circuit if posterior prediction is requested
    if Z is None:
        _posterior_predictive(Z_pred)
        return

    # Observation model based on the Nyquist plot
    sigma = {
        "real": numpyro.sample("sigma.real", dist.Exponential(rate=1.0)),
        "imag": numpyro.sample("sigma.imag", dist.Exponential(rate=1.0)),
    }
    numpyro.sample("obs.real", dist.Normal(Z_pred.real, sigma["real"]), obs=Z.real)
    numpyro.sample("obs.imag", dist.Normal(Z_pred.imag, sigma["imag"]), obs=Z.imag)


def circuit_regression_bode(
    freq: np.ndarray[float],
    priors: Mapping[str, Distribution],
    circuit_fn: Callable[[np.ndarray | float, np.ndarray], np.ndarray[complex]],
    Z: np.ndarray[complex] = None,
):
    """Inference model for ECM parameters based on Bode plot.

    This model infers the circuit parameters by matching the impedance
    measurements as plotted on the Bode plot. To do this, the MCMC sampling is
    performed on the magnitude and phase of the impedance.

    Parameters
    ----------
    freq : np.ndarray[float]
        Frequencies to evaluate the posterior predictive distribution at.
    priors : Mapping[str, Distribution]
        Priors for the circuit parameters as a dictionary of parameter labels
        and distributions.
    circuit_fn : Callable[[np.ndarray | float, np.ndarray], np.ndarray[complex]]
        Function to compute the circuit impedance, parameterized by frequency
        and circuit parameters, fn(freq, p).
    Z : np.ndarray[complex], optional
        Observed impedance data. If ``None``, the model will be used for
        posterior prediction, otherwise it will be used for Bayesian
        inference of circuit parameters.
    """
    # Make a deep copy of the priors to avoid side effects (pyro-ppl/numpyro/issues/1651)
    priors = copy.deepcopy(priors)
    # Sample each element of X separately
    p = jnp.array([numpyro.sample(k, v) for k, v in priors.items()])
    # Predict Z using the model
    Z_pred = circuit_fn(freq, p)
    mag, phase = jnp.abs(Z_pred), jnp.angle(Z_pred)

    # Custom observation model based on the Bode plot
    if Z is None:
        mag_gt = phase_gt = None
    else:
        mag_gt, phase_gt = jnp.abs(Z), jnp.angle(Z)
        # Log-transform the magnitude, otherwise low-frequency values dominate
        mag, mag_gt = jnp.log10(mag), jnp.log10(mag_gt)
    sigma_mag = numpyro.sample("sigma.mag", dist.Exponential(rate=1.0))
    sigma_phase = numpyro.sample("sigma.phase", dist.Exponential(rate=1.0))
    dist_mag = dist.TruncatedDistribution(dist.Normal(mag, sigma_mag), low=mag)
    dist_phase = dist.Normal(phase, sigma_phase)
    numpyro.sample("obs.mag", dist_mag, obs=mag_gt)
    numpyro.sample("obs.phase", dist_phase, obs=phase_gt)

    # NOTE: The following code is an alternative to the above code
    # error_mag = jnp.abs(jnp.log10(mag) - jnp.log10(mag_gt))
    # sigma_mag = numpyro.sample("sigma.mag", dist.HalfNormal())
    # numpyro.sample("obs.error.mag", dist.HalfNormal(sigma_mag), obs=error_mag)
    # error_phase = jnp.abs(phase - phase_gt)
    # sigma_phase = numpyro.sample("sigma.phase", dist.HalfNormal())
    # numpyro.sample("obs.error.phase", dist.HalfNormal(sigma_phase), obs=error_phase)


def _posterior_predictive(Z: np.ndarray[complex]):
    """Private helper to compute the posterior predictive distribution."""
    # NOTE: Z is predicted impedance, not observed impedance
    sigma_real = numpyro.sample("sigma.real", dist.Exponential(rate=1.0))
    sigma_imag = numpyro.sample("sigma.imag", dist.Exponential(rate=1.0))
    numpyro.sample("obs.real", dist.Normal(Z.real, sigma_real))
    numpyro.sample("obs.imag", dist.Normal(Z.imag, sigma_imag))
